ecs = { }

local ALL_ENTITIES = {}

---@alias System { callback: fun(delta: number, a: any, b: any, c: any, d: any, e: any, any), shapes: {}, keys: string[], entityCache: table<any, boolean> }

---@type System[]
local SYSTEMS = {}

local PLACEHOLDER = {}

-- TODO: Add entity to any existing systems
function ecs.addEntity(entity)
	ALL_ENTITIES[entity] = true
	for _, system in pairs(SYSTEMS) do
		if entityMatchesShapes(entity, system.shapes) then
		    system.entityCache[entity] = true
		end
	end
end

function ecs.removeEntity(entity)
	ALL_ENTITIES[entity] = nil
	for _, system in pairs(SYSTEMS) do
	    system.entityCache[entity] = nil
	end
end

---@return any
function ecs.field()
	return PLACEHOLDER
end

function allKeysIncluded(entity, filter)
	for k, _ in pairs(filter) do
		if not entity[k] then
		    return false
		end
	end
	return true
end

function entityMatchesShapes(entity, shapes)
	for _, shape in pairs(shapes) do
		if not allKeysIncluded(entity, shape) then
		    return false
		end
	end
	return true
end

---@generic T
---@generic U
---@generic V
---@generic W
---@param tShape T
---@param uShape U?
---@param vShape V?
---@param wShape W?
---@return fun(callback: fun(componentT: T, componentU: U, componentV: V, componentW: W))
function ecs.entitiesHavingShapes(tShape, uShape, vShape, wShape)
	return function()
	end
end

-- Print contents of `tbl`, with indentation.
-- `indent` sets the initial level of indentation.
function tprint (tbl, indent)
  if not indent then indent = 0 end
  for k, v in pairs(tbl) do
    formatting = string.rep("  ", indent) .. k .. ": "
    if type(v) == "table" then
      print(formatting)
      tprint(v, indent+1)
    elseif type(v) == 'boolean' then
      print(formatting .. tostring(v))
    else
      print(formatting .. v)
    end
  end
end

function addSystem(callback, keys, shapes)
	SYSTEMS[#SYSTEMS + 1] = {
		callback = callback,
		keys = keys,
		shapes = shapes,
		entityCache = nil
	}
end

---@param deltaSeconds number
function ecs.update(deltaSeconds)
	for _,system in pairs(SYSTEMS) do
		if not system.entityCache then
		    system.entityCache = {}
			for entity,_ in pairs(ALL_ENTITIES) do
				if entityMatchesShapes(entity, system.shapes) then
					system.entityCache[entity] = true
				end
			end
		end
		for entity,_ in pairs(system.entityCache) do
			local keys = system.keys
			system.callback(deltaSeconds, entity[keys[1]], entity[keys[2]], entity[keys[3]], entity[keys[4]], entity)
		end
	end
end

--- Returns a function that accepts a callback. This callback will receive one argument for each Shape provided.
---
---@generic T
---@generic TKey
---@generic U
---@generic UKey
---@generic V
---@generic VKey
---@generic W
---@generic WKey
---@param tShape { [TKey]: T }
---@param uShape { [UKey]: U } | fun(componentT: T, any) | nil
---@param vShape { [VKey]: V } | fun(componentT: T, componentU: U, any) | nil
---@param wShape { [WKey]: W } | fun(componentT: T, componentU: U, componentV: V, any) | nil
---@param finalFunc fun(componentT: T, componentU: U, componentV: V, componentW: W, any) | nil
function ecs.entitiesWithShapes(tShape, uShape, vShape, wShape, finalFunc)
	local maybeShapes = {tShape, uShape, vShape, wShape, finalFunc}
	local shapes = {}
	local callback

	for _,maybeShape in pairs(maybeShapes) do
		if type(maybeShape) == "table" then
		    shapes[#shapes + 1] = maybeShape
		elseif type(maybeShape) == "function" then
			callback = maybeShape
		end
	end

	local keys = {}
	for _,shape in pairs(shapes) do
		for key,_ in pairs(shape) do
			keys[#keys + 1] = key
		end
	end

	addSystem(callback, keys, shapes)
end

local f = ecs.field()

---@type { x: number, y: number }
local XYPair = { x = f, y = f }

local Position = { position = XYPair }
local Velocity = { velocity = XYPair }

local someMovable = {
	position = { x = 0, y = 0 },
	velocity = { x = 2, y = 1 }
}

ecs.addEntity(someMovable)

ecs.entitiesWithShapes(Position, Velocity, function (delta, pos, vel)
	print("position:")
	tprint(pos, 1)
	pos.x = pos.x + (vel.x * delta)
	pos.y = pos.y + (vel.y * delta)
end)